/* A ZPU emulator
 *
 * Based on an original idea by Toby Seckshund.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include <termio.h>

#include "zpu_memory.h"
#include "zpu.h"


#define EMULATE 1

uint32_t pc;
uint32_t sp;
uint8_t  instruction;
uint32_t cpu;
uint32_t topOfStack;

bool touchedPc = false;
bool decodeMask = false;

uint32_t flip(uint32_t i)
{
    uint32_t t = 0;
    for (int j = 0; j < 32; j++)
    {
        t |= ((i >> j) & 1) << (31 - j);
    }
    return t;
}

uint32_t pop()
{
    uint32_t val;
    val = memoryReadLong(sp);
    sp = sp + 4;
    return (val);
}
void push(uint32_t val)
{
    sp = sp - 4;
    topOfStack = val;
    memoryWriteLong(sp, topOfStack);
}

void emulate()
{
    push(pc + 1);
    pc = (memoryReadByte(pc) - 32) * VECTORSIZE + VECTORBASE;
    touchedPc = true;
}


#define SYS_READ  4
#define SYS_WRITE 5

void syscall()
{    
    int returnAdd = memoryReadLong(sp + 0);
    int errNoAdd = memoryReadLong(sp + 4);
    int fileNo = memoryReadLong(sp + 12);
    int charIndex = memoryReadLong(sp + 16);
    int stringLength = memoryReadLong(sp + 20);
    int sysCallId = memoryReadLong(sp + 8);
    switch (sysCallId)
    {
	case SYS_WRITE:
	    for (int i = 0; i < stringLength; i++)
	    {
		putchar(memoryReadByte(charIndex++));
	    }	
	    // Return value via R0 (AKA memory address 0)
	    memoryWriteLong(0, stringLength);
	    break;
	case SYS_READ:
	    for (int i = 0; i < stringLength; i++)
	    {
		memoryWriteByte(charIndex++, tty_getchar());
	    }	
	    // Return value via R0 (AKA memory address 0)
	    memoryWriteLong(0, stringLength);
	    break;
    }    
}



static struct termio savemodes;
static int havemodes = 0;

int tty_break()
{
    struct termio modmodes;
    if(ioctl(fileno(stdin), TCGETA, &savemodes) < 0)
    {
      exit(1);
    }
    havemodes = 1;
    modmodes = savemodes;
    modmodes.c_lflag &= ~ICANON;
    modmodes.c_cc[VMIN] = 1;
    modmodes.c_cc[VTIME] = 0;
    return ioctl(fileno(stdin), TCSETAW, &modmodes);
}

int tty_getchar()
{
    return getchar();
}

int tty_fix()
{
    if(!havemodes)
    {
	return (0);
	return ioctl(fileno(stdin), TCSETAW, &savemodes);
    }
}

uint32_t revBytes(uint32_t val)
{
    uint8_t byte_0 = val         & 0x000000ff;
    uint8_t byte_1 = (val >> 8)  & 0x000000ff;
    uint8_t byte_2 = (val >> 16) & 0x000000ff;
    uint8_t byte_3 = (val >> 24) & 0x000000ff;

    return((byte_0 << 24) + (byte_1 << 16) + (byte_2 << 8) + (byte_3));
}



void zpu_reset()
{
    int i;
    if (tty_break() != 0)
    {
      return exit(1);
    }
//  tty_fix();
    memoryInitialize();
    zpu_load();
    pc = 0;
    touchedPc = true;
    sp = 0;
    topOfStack = 0;
}

void zpu_execute()
{
    static int step = 0;

    // Not necessary here but does what has to be done
    // for the stack op optimized version.
    //sp = sp - 4;
    sp = 0x1fff8;

    for (;;)
    {
        touchedPc = false;

        instruction = memoryReadByte(pc);
//#define SINGLE_STEP
#ifdef SINGLE_STEP
//	printf ("PC=%08x SP=%08x TOS=%08x OP=%02x DM=%02x debug=%08x\n", pc, sp, memoryReadLong(sp), instruction, decodeMask, 0);
	printf ("0x%07x 0x%02x 0x%08x 0x%08x 0x%08x\n", pc, instruction, sp, memoryReadLong(sp), memoryReadLong(sp + 4));
	fflush(0);	
	//memoryDisplayLong(sp - 16*4, 32);
	//printf("\n");
	//getchar();
	if (step++ == 10000)
	{
	    printf("Done.\n");
	    fflush(0);	
	    exit(0);
	}
#endif
	
        if ((instruction & 0x80) == ZPU_IM)
        {
            uint32_t t;
            uint32_t a;
            t = instruction & 0x7f;
            //t = t << 25;
	    //t = t >> 25;
            if (t & 0x40)
            {
                t = t | 0xffffff80;
            }
            if (decodeMask)
            {
                a = pop();
                a = a << 7;
                a = a | (t & 0x7F);
                push(a);
            }
            else
            {
                push(t);
            }
            decodeMask = true;
        }
        else
        {
            decodeMask = false;
            if ((instruction & 0xF0) == ZPU_ADDSP)
            {
                uint32_t offset;
                uint32_t valAddr;
                int32_t a;
                offset = instruction & 0x0F;
                valAddr = sp + offset * 4;
                a = pop();
                a = a + memoryReadLong(valAddr);
                push(a);
            }
            else if ((instruction & 0xE0) == ZPU_LOADSP)
            {
                uint32_t addr;
                uint32_t offset;
                uint32_t a;
                offset = (instruction & 0x1F) ^ 0x10; 
                addr = sp + 4 * offset;
                a = memoryReadLong(addr);
                push(a);
            }
            else if ((instruction & 0xE0) == ZPU_STORESP)
            {
                uint32_t addr;
                uint32_t offset;
                uint32_t a;
                offset = (instruction & 0x1F) ^ 0x10;
                addr = sp + 4 * offset;
                a = pop();
                memoryWriteLong(addr, a);
            }
            else
            {
                switch (instruction)
                {
                    case 0:
                        printf ("Breakpoint\n");
			printf ("PC=%08x SP=%08x  OP=%02x DM=%02x debug=%08x\n", pc, sp, instruction, decodeMask, 0);
                        exit(1);
			break;
                    case ZPU_PUSHPC:
                        push(pc);
                        break;
                    case ZPU_OR:
                    {
			uint32_t a;
			uint32_t b;
			a = pop();
			b = pop();
			a = a | b;				
                        push(a);
                        break;
                    }
                    case ZPU_NOT:
                    {
			uint32_t a;
			a = pop();
			a = ~a;
                        push(a);
                        break;
                    }
                    case ZPU_LOAD:
                    {
                        uint32_t addr;
			uint32_t val;
                        addr = pop();
                        val = memoryReadLong(addr);
                        //val = revBytes(val);
                        push(val);
                        break;
                    }
                    case ZPU_PUSHSPADD:
                    {
#ifdef EMULATE_PUSHSPADD
			emulate();
#else
                        uint32_t a;
                        uint32_t b;
                        a = sp;
                        b = pop();
                        b = (b * 4) + a;
                        push(b);
#endif
                        break;
                    }
                    case ZPU_STORE:
		    {
                        uint32_t addr;
                        uint32_t val;
                        addr = pop();
                        val = pop();
                        memoryWriteLong(addr, val);
                        break;
		    }
                    case ZPU_POPPC:
                    {
                        pc = pop();
                        touchedPc = true;
                        break;
                    }
                    case ZPU_POPPCREL:
                    {
#ifdef EMULATE_POPPCREL
			emulate();
#else
                        uint32_t a;
                        a = pop();
                        pc = pc + a;
                        touchedPc = true;
#endif
                        break;
                    }
                    case ZPU_FLIP:
                    {
                        uint32_t a;
                        a = pop();
                        a = flip(a);
                        push(a);
                        break;
                    }
                    case ZPU_ADD:
                    {
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        a = a + b;
                        push(a);
                        break;
                    }
                    case ZPU_SUB:
                    {
#ifdef EMULATE_SUB
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        a = b - a;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_PUSHSP:
                    {
                        uint32_t a;
                        a = sp;
                        push(a);
                        break;
                    }
                    case ZPU_POPSP:
                    {
                        uint32_t a;
                        a = pop();
                        sp = a;
                        break;
                    }
                    case ZPU_NOP:
                        break;
                    case ZPU_AND:
                    {
                        uint32_t a;
                        uint32_t b;
                        a = pop();
                        b = pop();
                        a = a & b;
                        push(a);
                        break;
                    }
                    case ZPU_XOR:
                    {
#ifdef EMULATE_XOR
			emulate();
#else
                        uint32_t a;
                        uint32_t b;
                        a = pop();
                        b = pop();
                        a = a ^ b;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_LOADB:
                    {
#ifdef EMULATE_LOADB
			emulate();
#else
                        uint32_t addr;
                        uint32_t val;
                        addr = pop();
                        val = memoryReadByte(addr);
                        push(val);
#endif
                        break;
                    }
                    case ZPU_STOREB:
                    {
#ifdef EMULATE_STOREB
			emulate();
#else
                        uint32_t addr;
                        uint32_t val;
                        addr = pop();
                        val = pop();
                        memoryWriteByte(addr, val);
#endif
                        break;
                    }
                    case ZPU_LOADH:
                    {
#ifdef EMULATE_LOADH
			emulate();
#else
                        uint32_t addr;
                        uint16_t val;
                        addr = pop();
                        val = memoryReadWord(addr);
                        push(val);
#endif
                        break;
                    }
                    case ZPU_STOREH:
                    {
#ifdef EMULATE_STOREH
			emulate();
#else
                        uint32_t addr;
                        uint16_t val;
                        addr = pop();
                        val = pop();
                        memoryWriteWord(addr, val);
#endif
                        break;
                    }
                    case ZPU_LESSTHAN:
                    {
#ifdef EMULATE_LESSTHAN
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        if (a < b)
                            a = 1;
                        else
                            a = 0;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_LESSTHANOREQUAL:
                    {
#ifdef EMULATE_LESSTHANOREQUAL
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        if (a <= b)
                            a = 1;
                        else
                            a = 0;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_ULESSTHAN:
                    {
#ifdef EMULATE_ULESSTHAN
			emulate();
#else
                        uint32_t a;
                        uint32_t b;
                        a = pop();
                        b = pop();
                        if (a < b)
                            a = 1;
                        else
                            a = 0;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_ULESSTHANOREQUAL:
                    {
#ifdef EMULATE_ULESSTHANOREQUAL
			emulate();
#else
                        uint32_t a;
                        uint32_t b;
                        a = pop();
                        b = pop();
                        if (a <= b)
                            a = 1;
                        else
                            a = 0;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_SWAP:
                    {
                        uint32_t a;
                        a = pop();
                        a = ((a >> 16) & 0xffff) | (a << 16);
                        push (a);
                        break;
                    }
                    case ZPU_MULT16X16:
                    {
                        uint32_t a;
                        uint32_t b;
                        a = pop();
                        b = pop();
                        a = (a & 0xffff) * (b & 0xffff);
                        push(a);
                        break;
                    }
                    case ZPU_EQBRANCH:
                    {
#ifdef EMULATE_EQUBRANCH
			emulate();
#else
                        uint32_t compare;
                        uint32_t target;
                        target = pop();
                        compare = pop();
                        if (compare == 0)
                        {
                            pc = pc + target;
                            touchedPc = true;
                        }
#endif
                        break;
                    }
                    case ZPU_NEQBRANCH:
                    {
#ifdef EMULATE_NEQBRANCH
			emulate();
#else
                        uint32_t compare;
                        uint32_t target;
                        target = pop();
                        compare = pop();
                        if (compare != 0)
                        {
                            pc = pc + target;
                            touchedPc = true;
                        }
#endif
                        break;
                    }
                    case ZPU_MULT:
                    {
#ifdef EMULATE_MULT
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        a = a * b;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_DIV:
                    {
#ifdef EMULATE_DIV
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        if (b == 0)
                        {
                            printf("Divide by zero\n");
                            exit(1);
                        }
                        a = a / b;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_MOD:
                    {
#ifdef EMULATE_MOD
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        if (b == 0)
                        {
                            printf ("Divide by zero\n");
                            exit(1);
                        }
                        a = a % b;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_LSHIFTRIGHT:
                    {
#ifdef EMULATE_LSHIFTRIGHT
			emulate();
#else
                        uint32_t shift;
                        uint32_t a;
                        shift = pop();
                        a = pop();
                        shift = shift & 0x3f;
                        a = a >> shift;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_ASHIFTLEFT:
                    {
#ifdef EMULATE_ASHIFTLEFT
			emulate();
#else
                        uint32_t shift;
                        int32_t a;
                        shift = pop();
                        a = pop();
                        shift = shift & 0x3f;
                        a = a << shift;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_ASHIFTRIGHT:
                    {
#ifdef EMULATE_ASHIFTRIGHT
			emulate();
#else
                        uint32_t shift;
                        int32_t a;
                        shift = pop();
                        a = pop();
                        shift = shift & 0x3f;
                        a = a >> shift;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_CALL:
                    {
#ifdef EMULATE_CALL
			emulate();
#else
                        uint32_t address;
                        address = pop();
                        push(pc + 1);
                        pc = address;
                        touchedPc = true;
#endif
                        break;
                    }
                    case ZPU_CALLPCREL:
                    {
#ifdef EMULATE_CALLPCREL
			emulate();
#else
                        uint32_t address;
                        address = pop();
                        push(pc + 1);
                        pc = pc + address;
                        touchedPc = true;
#endif
                        break;
                    }
                    case ZPU_EQ:
                    {
#ifdef EMULATE_EQ
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        if (a == b)
                            a = 1;
                        else
                            a = 0;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_NEQ:
                    {
#ifdef EMULATE_NEQ
			emulate();
#else
                        int32_t a;
                        int32_t b;
                        a = pop();
                        b = pop();
                        if (a != b)
                            a = 1;
                        else
                            a = 0;
                        push (a);
#endif
                        break;
                    }
                    case ZPU_NEG:
                    {
#ifdef EMULATE_NEG
			emulate();
#else
                        int32_t a;
                        a = pop();
                        a = -a;
                        push(a);
#endif
                        break;
                    }
                    case ZPU_CONFIG:
                    {
#ifdef EMULATE_CONFIG
			emulate();
#else
                        uint32_t a;
                        a = pop();
                        cpu = a;
                        printf ("CONFIG indicates CPU type is %d\n", cpu);
#endif
                        break;
		    }
                    case ZPU_SYSCALL:
			emulate();
			syscall();
                        break;
                    default:
                        printf ("Illegal Instruction\n");
                        exit(1);
                        break;
                }
            }
        }
        if (!touchedPc)
        {
            pc = pc + 1;
            touchedPc = true;
        }
    }
}
